from .dataset import DataexchangeDataset


def create_dataexchange_dataset(dataset_id, /, *, root_path):
    return DataexchangeDataset(dataset_id=dataset_id, root_path=root_path)


def download_dataexchange_dataset_revision(
    output_path, dataset_id, revision_id=None, /
):
    client = create_dataexchange_dataset(dataset_id, root_path=output_path)
    return client.download_revision(revision_id)



import os
import structlog


def get_tome_manifest_key_fs(path):
    return os.path.join(path, "tome")


def get_tome_path_fs(root_path, prefix, tome_name):
    tome_parent_path = root_path if prefix is None else os.path.join(root_path, prefix)
    return os.path.join(tome_parent_path, tome_name)


def get_page_key_fs(path, subtype, page):
    name = page[subtype]
    return os.path.join(path, name)


def filter_ds_reader_logs(_, __, event_dict):
    if event_dict.get("client") == "ds_reader_fs":
        raise structlog.DropEvent
    return event_dict



import os
import time
import urllib

import structlog
import boto3


class DataexchangeDataset:
    def __init__(self, *, root_path, prefix=None, dataset_id, log=None):
        self._log = log if log is not None else structlog.get_logger()
        self._log = self._log.bind(
            client="dataexchange_dataset",
            root_path=root_path,
            prefix=prefix,
            dataset_id=dataset_id,
        )
        self._client = boto3.client("dataexchange")
        self.dataset_id = dataset_id
        self.path = get_dataset_path(root_path, prefix)
        self.dataset_name = None
        self._dataset = None

    def download_revision(self, revision_id=None):
        self._init()
        rev_id = revision_id if revision_id is not None else self._get_latest_revision()
        self._log.info(
            "Download Revision: Start", revision_id=rev_id, output_path=self.path
        )

        assets = self._get_assets(rev_id)
        for asset in assets:
            self._download_asset(asset)

        self._log.info(
            "Download Revision: End", revision_id=rev_id, output_path=self.path
        )

    def _init(self):
        if self._dataset is not None:
            return

        res = self._client.get_data_set(DataSetId=self.dataset_id)
        self.dataset_name = res.get("Name")
        self.path = os.path.join(self.path, self.dataset_name)

    def _get_latest_revision(self):
        res = self._client.list_data_set_revisions(DataSetId=self.dataset_id)
        return res.get("Revisions")[0].get("Id")

    def _get_assets(self, revision_id):
        res = self._client.list_revision_assets(
            DataSetId=self.dataset_id, RevisionId=revision_id
        )
        return res.get("Assets")

    def _download_asset(self, asset):
        asset_name = asset.get("Name")
        output_path = os.path.join(self.path, asset_name)
        output_dir = os.path.dirname(output_path)
        os.makedirs(output_dir, exist_ok=True)

        if os.path.isfile(output_path):
            return

        url = self._get_download_asset_signed_url(asset)
        urllib.request.urlretrieve(url, output_path)
        self._log.info("Downloaded Asset", path=output_path, asset_name=asset_name)

    def _get_download_asset_signed_url(self, asset):
        job = self._client.create_job(
            Type="EXPORT_ASSET_TO_SIGNED_URL",
            Details={
                "ExportAssetToSignedUrl": {
                    "AssetId": asset.get("Id"),
                    "DataSetId": asset.get("DataSetId"),
                    "RevisionId": asset.get("RevisionId"),
                }
            },
        )

        job_id = job.get("Id")
        self._client.start_job(JobId=job_id)

        while True:
            time.sleep(1)
            job = self._client.get_job(JobId=job_id)

            if job.get("State") == "ERROR":
                message = job.get("Errors")[0].get("Message")
                raise Exception(f"Job {job_id} failed to complete: {message}")

            if job.get("State") == "COMPLETED":
                return job.get("Details").get("ExportAssetToSignedUrl").get("SignedUrl")


def get_dataset_path(root_path, prefix):
    parent_path = root_path if prefix is None else os.path.join(root_path, prefix)
    return parent_path



import os

import structlog
import rapidjson
import pandas as pd

from .constants import (
    get_tome_manifest_key_fs,
    get_page_key_fs,
    get_tome_path_fs,
)


class TomeWriterFs:
    def __init__(self, *, root_path, prefix=None, tome_name, log=None):
        self._log = log if log is not None else structlog.get_logger()
        self._log = self._log.bind(
            client="tome_writer_fs",
            root_path=root_path,
            prefix=prefix,
            tome_name=tome_name,
        )
        self.tome_name = tome_name
        self.path = get_tome_path_fs(root_path, prefix, tome_name)
        self._parquet_compression = "gzip"

    def write_manifest(self, manifest):
        self._ensure_dir()
        self._log.info("Write Manifest: Start")
        key = get_tome_manifest_key_fs(self.path)
        self._write_json(key, manifest)

    def write_page(self, page, dataframe, keyset):
        self._ensure_dir()
        self._log.info("Write Page Start", page_number=page["number"])
        self._write_dataframe(page, dataframe)
        self._write_keyset(page, keyset)

    def _ensure_dir(self):
        if not os.path.isdir(self.path):
            os.makedirs(self.path)

    def _write_dataframe(self, page, dataframe):
        key = self._get_page_key("dataframe", page)

        content_type = page["dataframeContentType"]
        if content_type != "application/x-parquet":
            raise Exception(f"Unsupported content type {content_type}")

        self._log.info("Write Dataframe: Start", page_number=page["number"])
        self._write_parquet(key, dataframe)

    def _write_keyset(self, page, keyset):
        key = self._get_page_key("keyset", page)

        content_type = page["keysetContentType"]
        if content_type != "application/x-parquet":
            raise Exception(f"Unsupported content type {content_type}")

        self._log.info("Write keyset: Start", page_number=page["number"])
        df = pd.DataFrame(
            keyset, columns=["_"]
        )  # must have string column name for parquet
        self._write_parquet(key, df)

    def _get_page_key(self, subtype, page):
        return get_page_key_fs(self.path, subtype, page)

    def _write_parquet(self, key: str, df: pd.DataFrame) -> None:
        self._log.debug("Write parquet", key=key)
        df.to_parquet(key, compression=self._parquet_compression)

    # pylint: disable=no-self-use
    def _write_json(self, key, data):
        with open(key, "w", encoding="utf-8") as file:
            rapidjson.dump(data, file)



import os
from glob import glob
import structlog

from ..ds_io import DsReaderFs, GameDsLoader
from .loader import TomeLoader
from .scribe import TomeScribe
from .manifest import TomeManifest
from .writer_fs import TomeWriterFs
from .reader_fs import TomeReaderFs
from .constants import filter_ds_reader_logs


def default_tome_name():
    return "header"


def create_header_tome_from_fs(
    tome_name=None,
    /,
    *,
    ds_type="csds",
    tome_collection_root_path="tomes",
    ds_collection_root_path="data",
    path_depth: int = 4,
    log=None,
):
    """Make the header tome"""
    name = default_tome_name() if tome_name is None else tome_name
    log = (
        log
        if log is not None
        else structlog.wrap_logger(
            structlog.get_logger(), processors=[filter_ds_reader_logs]
        )
    )

    writer = TomeWriterFs(root_path=tome_collection_root_path, tome_name=name, log=log)
    manifest = TomeManifest(
        tome_name=name, path=writer.path, ds_type=ds_type, is_header=True
    )
    scribe = TomeScribe(manifest=manifest, writer=writer, log=log)

    scribe.start()

    for ds_path in get_ds_paths_from_glob(ds_collection_root_path, path_depth, ds_type):
        ds_loader = fetch_ds_header_from_fs(ds_path, ds_type, log)
        job_id = ds_loader.manifest["jobId"]
        df = ds_loader.get_channel({"channel": "header"})
        df["ds_path"] = ds_path
        df["key"] = job_id
        scribe.concat(df, job_id)

    scribe.finish()

    reader = TomeReaderFs(root_path=tome_collection_root_path, tome_name=name, log=log)

    return TomeLoader(reader=reader, log=log)


# pylint: disable=too-many-locals
def create_subheader_tome_from_fs(
    name,
    /,
    src_tome_name=None,
    selector=lambda df: [True] * len(df),
    *,
    tome_collection_root_path="tomes",
    dest_tome_name=None,
    preserve_src_id=False,
    log=None,
):
    """
    selector is passed to filter out rows from the header
    by doing df = df.loc[selector]
    """
    log = log if log is not None else structlog.get_logger()
    src_name = default_tome_name() if src_tome_name is None else src_tome_name

    src_reader = TomeReaderFs(
        root_path=tome_collection_root_path, tome_name=src_name, log=log
    )
    src_loader = TomeLoader(reader=src_reader, log=log)

    if dest_tome_name is None:
        output_path = tome_collection_root_path
    else:
        output_path = os.path.join(tome_collection_root_path, dest_tome_name)

    writer = TomeWriterFs(root_path=output_path, tome_name=name, log=log)
    manifest = TomeManifest(
        tome_name=name,
        path=writer.path,
        ds_type=src_loader.manifest["dsType"],
        header_tome_name=src_name,
        is_header=True,
        src_id=src_loader.manifest["id"] if preserve_src_id else None,
    )
    scribe = TomeScribe(writer=writer, manifest=manifest, log=log)
    scribe.start()
    df = src_loader.get_dataframe().loc[selector]
    scribe.concat(df, list(df["key"]))
    scribe.finish()

    reader = TomeReaderFs(root_path=output_path, tome_name=name, log=log)

    return TomeLoader(reader=reader, log=log)


def get_ds_paths_from_glob(ds_root_path, path_depth, ds_type):
    paths = glob(os.path.join(ds_root_path, *(["*"] * path_depth), ds_type))
    paths = [os.path.join(*path.split(os.path.sep)[:-1]) for path in paths]
    paths = set(paths)
    return paths


def fetch_ds_header_from_fs(ds_path, ds_type, log):
    root_path = os.path.join(*ds_path.split(os.path.sep)[:-1])
    ds_key = ds_path.split(os.path.sep)[-1]
    ds_reader = DsReaderFs(
        root_path=root_path,
        manifest_key=os.path.join(ds_key, ds_type),
        log=log,
    )
    ds_loader = GameDsLoader(reader=ds_reader, log=log)
    return ds_loader



"""
Python Data Science Development Kit.
"""
from .ds_models import create_ds_models
from .ds_io.reader_s3 import DsReaderS3
from .ds_io.reader_fs import DsReaderFs
from .ds_io.game_ds_loader import GameDsLoader
from .tome import TomeCuratorFs, create_tome_curator
from .exchange import (
    create_dataexchange_dataset,
    download_dataexchange_dataset_revision,
)



import pandas as pd
import structlog
import rapidjson
from .constants import (
    get_page_key_fs,
    get_tome_manifest_key_fs,
    get_tome_path_fs,
)


class TomeReaderFs:
    def __init__(self, *, root_path, prefix=None, tome_name, log=None, has_header=True):
        self._log = log if log is not None else structlog.get_logger()
        self._log = self._log.bind(
            client="tome_reader_fs",
            root_path=root_path,
            prefix=prefix,
            tome_name=tome_name,
        )

        self._path = get_tome_path_fs(root_path, prefix, tome_name)
        self.has_header = has_header
        self.header = None
        if self.has_header:
            self.header = TomeReaderFs(
                root_path=self._path,
                tome_name="header",
                has_header=False,
                log=self._log,
            )

    @property
    def exists(self):
        """If the tome exists"""
        try:
            self.read_manifest()
        except FileNotFoundError:
            return False
        except:
            self._log.error("There was an error while loading the loader")
            raise
        return True

    def read_manifest(self):
        key = get_tome_manifest_key_fs(self._path)
        with open(key, "r", encoding="utf-8") as file:
            data = rapidjson.loads(file.read())
        return data

    # pylint: disable=no-self-use
    def read_metadata(self):
        return {}

    def read_page(self, page):
        dataframe = self.read_page_dataframe(page)
        keyset = self.read_page_keyset(page)
        return dataframe, keyset

    def read_page_keyset(self, page):
        key = self._get_page_key("keyset", page)

        content_type = page["keysetContentType"]
        if content_type != "application/x-parquet":
            raise Exception(f"Unknown content type {content_type}")

        self._log.info("Read keyset: Start", page_number=page["number"])
        df = pd.read_parquet(key)
        return list(df.iloc[:, 0])

    def read_page_dataframe(self, page):
        key = self._get_page_key("dataframe", page)

        content_type = page["dataframeContentType"]
        if content_type != "application/x-parquet":
            raise Exception(f"Unsupported content type {content_type}")

        self._log.info("Read Dataframe: Start", page_number=page["number"])
        df = pd.read_parquet(key)
        return df

    def _get_page_key(self, subtype, page):
        return get_page_key_fs(self._path, subtype, page)
